import numpy as np
import hex3_old as hex3
import hex3_chen_old as hex3_chen
import copy
import pandas as pd


def gen_dataset(N, ttratio=1.0):
    # Smith, Noah A., and Roy W. Tromble. "Sampling uniformly from the unit simplex." Johns Hopkins University, Tech. Rep 29 (2004).
    dim = 3
    x = np.sort(np.random.rand(dim - 1, N * 10), axis=0)
    x = np.concatenate([np.zeros((1, N * 10)), x, np.ones((1, N * 10))], axis=0)
    alpha = x[1:] - x[:-1]

    # # Checking uniformity
    # ax = plt.axes(projection='3d')
    # ax.plot(alpha[0], alpha[1], alpha[2] ,'b.')
    # plt.show()

    alpha = alpha[:-1]

    parspan = {}
    # Defining disturbance box [center, variability]
    parspan['T0'] = [60, 10]  # C
    parspan['w0'] = [105, 25]  # kW/K
    parspan['wh1'] = [40, 10]  # kW/K
    parspan['wh2'] = [50, 10]  # kW/K
    parspan['wh3'] = [30, 10]  # kW/K
    parspan['Th1'] = [150, 30]  # C
    parspan['Th2'] = [150, 30]  # C
    parspan['Th3'] = [150, 30]  # C
    parspan['UA1'] = [65, 15]  # kW/K
    parspan['UA2'] = [80, 10]  # kW/K
    parspan['UA3'] = [95, 15]  # kW/K

    # Copied from transfer learning
    parspan['Ts'] = [0, 0]  # C
    parspan['h1'] = [0, 0]  # kW/K
    parspan['h2'] = [0, 0]  # kW/K
    parspan['h3'] = [0, 0]  # kW/K

    randmatrix = np.random.rand(len(parspan), N * 10)
    parvec = {}
    for i, parname in enumerate(parspan.keys()):
        parvec[parname] = parspan[parname][0] + ttratio * (2 * randmatrix[i] - 1) * (parspan[parname][-1])

    par0 = [{key: value[i] for key, value in parvec.items()} for i in range(N * 10)]

    # Generating measurements, priors and targets
    u_span = []
    u_rand_span = []
    u_span_chen = []
    d_span = []
    J_span = []
    J_span_chen = []

    hex3_chen.Ti_max = 135
    hex3.Ti_max = 135

    errors = 0
    finished = 0
    i = 0
    while finished < N:
        params = par0[i]

        if any(alpha[:, i] < 0.05) or np.sum(alpha[:, i]) > 0.95:
            errors += 1
            i += 1
            print('Too low or high alphas: ', errors)
            continue

        u_chen = hex3_chen.optim(copy.deepcopy(params))
        # Note: hex3 model data not used for predictions
        u = hex3.optim(copy.deepcopy(params))
        # Calculate optimal output temp from optimal u
        if not u_chen['success'] or not u['success']:
            errors += 1
            i += 1
            print('Bad u_opt, errors: ', errors)
            continue

        cost = hex3.cost(u['u'], copy.deepcopy(params))
        cost_chen = hex3_chen.cost(u_chen['u'], copy.deepcopy(params))

        if not cost['success'] or not cost_chen['success']:
            errors += 1
            i += 1
            print('Bad hex cost solved, errors: ', errors)
            continue
        else:
            print('Success solutions: ', finished + 1)
            i += 1
            finished += 1

        # print(cost)
        # Save values
        u_span.append(np.array(u['u']))
        u_rand_span.append(alpha[:, i])
        u_span_chen.append(np.array(u_chen['u']))
        d_span.append(params)
        J_span.append(-cost['J'][0])
        J_span_chen.append(-cost_chen['J'][0])

        # For Scipy Implemntation (NOT USED)
        # params = par0[i]
        #
        # u = hex3_chen.optim(copy.deepcopy(params))
        # # Calculate optimal output temp from optimal u
        # if not u['success']:
        #     errors += 1
        #     i+=1
        #     print('Bad u_opt, errors: ', errors)
        #     continue
        # cost = hex3.cost(u['u'], copy.deepcopy(params))
        # cost_chen = hex3_chen.cost(u['u'], copy.deepcopy(params))
        # # gradient_chen = hex3_chen.grad(u, copy.deepcopy(params))
        # # grad = hex3.grad(u, copy.deepcopy(params))['grad']
        #
        # if not cost['success']:
        #     errors += 1
        #     i+=1
        #     print('Bad hex cost solved, errors: ', errors)
        #     continue
        # elif not cost_chen['success']:
        #     errors += 1
        #     i += 1
        #     print('Bad hex chen cost solved, errors: ', errors)
        #     continue
        # else:
        #     print('Success solutions: ', finished)
        #     i += 1
        #     finished += 1

        # # Save values
        # u_span.append(np.array(u['u']))
        # d_span.append(params)
        # J_span.append(-cost['J'])
        # J_span_chen.append(-cost_chen['J'])

    u_span = np.array(u_span)
    u_rand_span = np.array(u_rand_span)
    u_span_chen = np.array(u_span_chen)
    d_span = np.array(d_span, dtype=dict)
    J_span = np.array(J_span)
    J_span_chen = np.array(J_span_chen)

    return u_span, u_span_chen, u_rand_span, d_span, J_span, J_span_chen


def save_data(name, u_span, u_span_chen, u_rand_span, d_span, J_span, J_span_chen):
    u_headers = [f'u{i}' for i in range(u_span.shape[1])]
    u_chen_headers = [f'uc{i}' for i in range(u_span_chen.shape[1])]
    u_rand_headers = [f'ur{i}' for i in range(u_rand_span.shape[1])]
    J_header = 'J'
    J_chen_header = 'J_chen'

    u_span_pd = pd.DataFrame.from_dict({key: val for key, val in zip(u_headers, u_span.T)})
    u_span_chen_pd = pd.DataFrame.from_dict({key: val for key, val in zip(u_chen_headers, u_span_chen.T)})
    u_rand_span_pd = pd.DataFrame.from_dict({key: val for key, val in zip(u_rand_headers, u_rand_span.T)})
    J_span_pd = pd.DataFrame.from_dict({J_header: J_span})
    # print(J_span)
    # print(J_span_chen)
    J_span_chen_pd = pd.DataFrame.from_dict({J_chen_header: J_span_chen})
    d_span_pd = pd.DataFrame.from_records(d_span)

    frames = pd.concat([u_span_pd, u_span_chen_pd, u_rand_span_pd, d_span_pd, J_span_pd, J_span_chen_pd], axis=1)

    frames.to_csv(name, index=False)


def load_data(name):
    frames = pd.read_csv(name)
    data = dict(u=frames.iloc[:, :2],
                u_chen=frames.iloc[:, 2:4],
                u_rand=frames.iloc[:, 4:6],
                d=frames.iloc[:, 6:-2],
                J=frames.iloc[:, -2:-1],
                J_chen=frames.iloc[:, -1:])
    return data


if __name__ == '__main__':

    # training sets
    for samples in [100, 500, 1000, 2500]:
        print('Generating training data....')

        np.random.seed(2030)
        u_span, u_span_chen, u_rand_span, d_span, J_span, J_span_chen = gen_dataset(samples, 1)
        save_data(f'.\\datasets\\train_u_constrained_prediction{samples}.csv', u_span, u_span_chen, u_rand_span, d_span, J_span,
                  J_span_chen)
        print('Done')

    # Test set
    print('Generating test data....')
    np.random.seed(2028)
    samples = 2500
    u_span, u_span_chen, u_rand_span, d_span, J_span, J_span_chen = gen_dataset(samples, 1.2)
    save_data(f'.\\datasets\\test_u_constrained_prediction{samples}.csv', u_span, u_span_chen, u_rand_span, d_span, J_span,
              J_span_chen)
    print('Done')
